
/* Chrysalide - Outil d'analyse de fichiers binaires
 * menubar.c - gestion des différents menus de la fenêtre principale
 *
 * Copyright (C) 2011-2017 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "menubar.h"


#include "binary.h"
#include "debug.h"
#include "edition.h"
#include "file.h"
#include "help.h"
#include "plugins.h"
#include "project.h"
#include "view.h"
#include "../editem-int.h"



/* Barre de menus de la fenêtre principale (instance) */
struct _GMenuBar
{
    GEditorItem parent;                     /* A laisser en premier        */

    GtkWidget *file;                        /* Menu "Fichier"              */
    GtkWidget *edition;                     /* Menu "Edition"              */
    GtkWidget *view;                        /* Menu "Affichage"            */
    GtkWidget *project;                     /* Menu "Projet"               */
    GtkWidget *binary;                      /* Menu "Binaire"              */
    GtkWidget *debug;                       /* Menu "Débogage"             */
    GtkWidget *plugins;                     /* Menu "Greffons"             */
    GtkWidget *help;                        /* Menu "Aide"                 */

};


/* Barre de menus de la fenêtre principale (classe) */
struct _GMenuBarClass
{
    GEditorItemClass parent;                /* A laisser en premier        */

};


/* Initialise la classe de la barre de menus de l'éditeur. */
static void g_menu_bar_class_init(GMenuBarClass *);

/* Initialise une instance de la barre de menus pour l'éditeur. */
static void g_menu_bar_init(GMenuBar *);

/* Supprime toutes les références externes. */
static void g_menu_bar_dispose(GMenuBar *);

/* Procède à la libération totale de la mémoire. */
static void g_menu_bar_finalize(GMenuBar *);

/* Lance une actualisation du fait d'un changement de vue. */
static void update_menu_bar_for_view(GMenuBar *, GtkDisplayPanel *);

/* Réagit à un changement de focus des panneaux d'affichage. */
static void notify_focus_change_for_menu_bar(GMenuBar *, GtkDisplayPanel *);

/* Met à jour les accès aux menus en fonction de la position. */
static void track_caret_address_for_menu_bar(GMenuBar *, GtkDisplayPanel *, const vmpa2t *);

/* Lance une actualisation relative à l'étendue du projet. */
static void update_menu_bar_for_project(GMenuBar *, GStudyProject *);



/* Indique le type défini pour la barre de menus de la fenêtre principale. */
G_DEFINE_TYPE(GMenuBar, g_menu_bar, G_TYPE_EDITOR_ITEM);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe de la barre de menus de l'éditeur.      *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_menu_bar_class_init(GMenuBarClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GEditorItemClass *editem;               /* Encore une autre vision...  */

    object = G_OBJECT_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_menu_bar_dispose;
    object->finalize = (GObjectFinalizeFunc)g_menu_bar_finalize;

    editem = G_EDITOR_ITEM_CLASS(klass);

    editem->update_view = (update_item_view_fc)update_menu_bar_for_view;
    editem->notify_focus = (notify_focus_change_fc)notify_focus_change_for_menu_bar;
    editem->track_caret = (track_caret_in_view_fc)track_caret_address_for_menu_bar;
    editem->update_project = (update_project_fc)update_menu_bar_for_project;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : bar = instance à initialiser.                                *
*                                                                             *
*  Description : Initialise une instance de la barre de menus pour l'éditeur. *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_menu_bar_init(GMenuBar *bar)
{
    GEditorItem *item;                      /* Autre version de l'élément  */

    item = G_EDITOR_ITEM(bar);

    item->name = "menus";

    item->widget = gtk_menu_bar_new();
    gtk_widget_show(item->widget);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : bar = instance d'objet GLib à traiter.                       *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_menu_bar_dispose(GMenuBar *bar)
{
    G_OBJECT_CLASS(g_menu_bar_parent_class)->dispose(G_OBJECT(bar));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : bar = instance d'objet GLib à traiter.                       *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_menu_bar_finalize(GMenuBar *bar)
{
    G_OBJECT_CLASS(g_menu_bar_parent_class)->finalize(G_OBJECT(bar));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : ref      = espace de référencement global.                   *
*                accgroup = groupe d'accélérateurs pour les menus.            *
*                                                                             *
*  Description : Compose la barre de menus principale.                        *
*                                                                             *
*  Retour      : Adresse de la structure mise en place.                       *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GEditorItem *g_menu_bar_new(GObject *ref, GtkAccelGroup *accgroup)
{
    GMenuBar *result;                       /* Structure à retourner       */
    GEditorItem *item;                      /* Autre version de l'élément  */

    result = g_object_new(G_TYPE_MENU_BAR, NULL);

    g_object_set_data(G_OBJECT(result), "accgroup", accgroup);

    /* Initialisation générique */

    item = G_EDITOR_ITEM(result);

    g_object_ref(ref);
    item->ref = ref;

    /* Fichier */

    result->file = build_menu_file(ref, accgroup);
    gtk_container_add(GTK_CONTAINER(item->widget), result->file);

    /* Edition */

    result->edition = build_menu_edition(ref, accgroup, result);
    gtk_container_add(GTK_CONTAINER(item->widget), result->edition);

    /* Affichage */

    result->view = build_menu_view(ref, accgroup, result);
    gtk_container_add(GTK_CONTAINER(item->widget), result->view);

    /* Projet */

    result->project = build_menu_project(ref, accgroup, result);
    gtk_container_add(GTK_CONTAINER(item->widget), result->project);

    /* Binaire */

    result->binary = build_menu_binary(ref, accgroup, result);
    gtk_container_add(GTK_CONTAINER(item->widget), result->binary);

    /* Débogage */

    result->debug = build_menu_debug(ref, accgroup);
    gtk_container_add(GTK_CONTAINER(item->widget), result->debug);

    /* Greffons */

    result->plugins = build_menu_plugins(ref, accgroup);
    gtk_container_add(GTK_CONTAINER(item->widget), result->plugins);

    /* Aide */

    result->help = build_menu_help(ref, accgroup);
    gtk_container_add(GTK_CONTAINER(item->widget), result->help);

    /* Finalisation générique */

    notify_focus_change_for_menu_bar(result, NULL);

    return G_EDITOR_ITEM(result);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : bar   = barre de menus à actualiser.                         *
*                panel = nouveau panneau d'affichage actif.                   *
*                                                                             *
*  Description : Lance une actualisation du fait d'un changement de vue.      *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void update_menu_bar_for_view(GMenuBar *bar, GtkDisplayPanel *panel)
{
    update_menu_view_for_view(bar->view, panel, bar);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : bar  = barre de menus à actualiser.                          *
*                panel = composant d'affichage concerné par l'opération.      *
*                                                                             *
*  Description : Réagit à un changement de focus des panneaux d'affichage.    *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void notify_focus_change_for_menu_bar(GMenuBar *bar, GtkDisplayPanel *panel)
{
    GBufferLine *line;                      /* Ligne de position courante  */
    const mrange_t *range;                  /* Couverture en mémoire       */
    const vmpa2t *addr;                     /* Position courante           */
    GEditorItem *item;                      /* Autre version de l'élément  */

    if (panel != NULL && gtk_display_panel_get_position(panel, &line, NULL))
    {
        range = g_buffer_line_get_range(line);
        addr = get_mrange_addr(range);
    }
    else
    {
        line = NULL;
        addr = NULL;
    }

    item = G_EDITOR_ITEM(bar);

    update_access_in_menu_edition(item->ref, panel, addr);

    update_access_in_menu_view(item->ref, panel);

    update_access_in_menu_binary(item->ref, panel);

    if (line != NULL)
        g_object_unref(G_OBJECT(line));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : bar   = barre de menus à actualiser.                         *
*                panel = panneau d'affichage actif ou NULL si aucun.          *
*                addr  = nouvelle adresse du curseur courant.                 *
*                                                                             *
*  Description : Met à jour les accès aux menus en fonction de la position.   *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void track_caret_address_for_menu_bar(GMenuBar *bar, GtkDisplayPanel *panel, const vmpa2t *addr)
{
    GEditorItem *item;                      /* Autre version de l'élément  */

    item = G_EDITOR_ITEM(bar);

    update_access_in_menu_edition(item->ref, panel, addr);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : bar     = barre de menus à actualiser.                       *
*                project = projet visé par la procédure.                      *
*                                                                             *
*  Description : Lance une actualisation relative à l'étendue du projet.      *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void update_menu_bar_for_project(GMenuBar *bar, GStudyProject *project)
{
    update_menu_project_for_project(bar->project, project, bar);

}
